import base64
import binascii
import datetime
import getpass
import ipaddress
import os.path
import re
import secrets
import sys
from enum import IntEnum
from typing import Optional, cast, Protocol
from urllib.parse import urlparse

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509.oid import ExtensionOID, NameOID, AuthorityInformationAccessOID
from cryptography.exceptions import InvalidSignature

"""
The RegEx to match a domain. Examples :
  - wiki
  - wiki.internal-domain.lan
  - *.sub.internal-domain.lan
"""
regex_domain = re.compile(r'^(\*\.)?([a-zA-Z0-9_-]+\.?)+$')

"""
The RegEx to match a Subject Alternate Name.

  - DNS: host, DNS: host2.domain.lan, IP: 10.0.0.1, IP: 2001:0db8:85a3:0000:0000:8a2e:0370:7334
"""
regex_san = re.compile(r'^(('
                       r'(DNS:[ ]?((\*\.)?([a-zA-Z0-9_-]+\.?))+)'  # DNS, see regex_domain
                       r'|'
                       r'(IP:[ ]?('  # IPv4
                       r'((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))'
                       r'|'  # or IPv6
                       r'((?:[a-fA-F0-9]{1,4}:){7}[a-fA-F0-9]{1,4})'
                       r')))'
                       r'(,?[ ]?))+'  # Followed by "," or ", "
                       r'(?<![ ,])$')  # But the string cannot be ending with "," or ", "

"""
The RegEx for the Common Name : host or IPv4 or IPv6
"""
regex_cn = re.compile(r'^(((\*\.)?([a-zA-Z0-9_-]+\.?)+)|'
                      r'((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))|'
                      r'((?:[a-fA-F0-9]{1,4}:){7}[a-fA-F0-9]{1,4}))$')


def ask(prompt: str, values: Optional[list] = None, disable_show_values: bool = False, can_be_empty: bool = False,
        disable_lowering: bool = False) -> Optional[str]:
    """
    Ask the user for an input.

    The prompt is written to stdout and the value is read from stdin.

    The prompt will be like that :
        "<prompt message> : "
        "<prompt message> (<value 1>, <value 2>, <value 3>) : "

    The optional disable_show_values parameter disable showing values in the prompt.

    The optional values parameter is the fixed list of the authorized values. If defined, the prompt is asking until the
    user is entering one of these values.

    The optional can_be_empty parameter allows the user to not enter a value but send back a empty value.

    By default, the input string is lower(). If you doesn't want that, you can use the disable_lowering parameter.

    :param prompt: the prompt message
    :param values: the list of valid values
    :param disable_show_values: disable showing values
    :param can_be_empty: indicate if the response can be empty
    :param disable_lowering: disable the lower() of the response
    :return: the user response
    """
    if values:
        values = [str(v).lower() for v in values]

        if not disable_show_values:
            prompt += f' ({", ".join(values)})'

    prompt += ' : '

    while True:
        print(prompt, end='')
        answer = input().strip()
        answer_lower = answer.lower()

        if values:
            if answer_lower in values or (can_be_empty and len(answer) == 0):
                break
        else:
            if (len(answer) == 0 and can_be_empty) or len(answer) > 0:
                break

    if len(answer) == 0:
        return None

    if disable_lowering:
        return answer
    else:
        return answer_lower

class PEMFormatType(IntEnum):
    """
    The PEM type
    """
    CERT = 1
    KEY = 2
    CSR = 3

def ask_pem_format(prompt: str, pem_type: PEMFormatType) -> Optional[str]:
    """
    Ask the user for data in the PEM format.

    The targeted data is providing through the pem_type parameter.

    This function accepts multi-line PEM. The first line must be the ASCII armored definition "----- BEGIN...". This
    function will stop reading from stdin when the the "----- END..." is reached.

    The PEM data is not checked.

    :param prompt: the prompt message
    :param pem_type: the PEM format type
    :return: the entered PEM data
    """
    answer = ''

    prompt += ' :'
    print(prompt)

    if pem_type == PEMFormatType.CERT:
        re_start_line = re.compile('-----BEGIN CERTIFICATE-----')
        re_end_line = re.compile('-----END CERTIFICATE-----')
    elif pem_type == PEMFormatType.KEY:
        re_start_line = re.compile('-----BEGIN (RSA )?PRIVATE KEY-----')
        re_end_line = re.compile('-----END (RSA )?PRIVATE KEY-----')
    elif pem_type.CSR:
        re_start_line = re.compile('-----BEGIN CERTIFICATE REQUEST-----')
        re_end_line = re.compile('-----END CERTIFICATE REQUEST-----')
    else:
        raise RuntimeError(f'PEM type {pem_type} not supported')

    end_line_detected = False
    first = True
    while 1:
        try:
            line = sys.stdin.readline()
        except KeyboardInterrupt:
            break

        if first:
            first = False
            if not re_start_line.match(line.rstrip()):
                raise ValueError('No or bad start line')

        answer += line

        if re_end_line.match(line.rstrip()):
            end_line_detected = True
            break

    if not end_line_detected:
        raise ValueError('No end line detected')

    return answer

def ask_date_in_future() -> datetime.date:
    """
    Ask the user for a date in the future.

    If an error is detected, the error is written on stdout and the same question is prompted again.

    :return: the date
    """
    error = True
    d = ''

    while error:
        d = ask('Enter a date in the future (DD/MM/YYYY)')

        try:
            d = datetime.datetime.strptime(d, '%d/%m/%Y').date()

            if d < datetime.date.today():
                print(f'\tThe date must be in the future')
            else:
                error = False
        except ValueError:
            print(f'\t{d} is invalid')

    return d

def ask_cn() -> str:
    """
    Ask the user for Common Name.

    The user can enter 1 host or IP.

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the CN
    """
    error = True
    cn = ''

    while error:
        cn = ask('Common Name')

        if not regex_cn.match(cn):
            print(f'\t{cn} is invalid')
        else:
            error = False

    return cn

def ask_hosts() -> list[str]:
    """
    Ask the user for hosts.

    The user can enter 0, 1 or n hosts separated by a comma. The host value can be :
        * the hostname : wiki
        * the FQDN : wiki.internal-domain.lan
        * the * wildcard : *.sub.internal-domain.lan

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the hosts
    """
    error = True
    domains = set()

    while error:
        input_domains = ask('Enter alternative names separated by a comma (eg: srv.internal.lan)', can_be_empty=True)
        domains.clear()

        if not input_domains or len(input_domains) == 0:
            break

        error = False
        for d in input_domains.split(','):
            d = d.strip()

            if regex_domain.match(d):
                try:
                    ipaddress.ip_address(d)
                    error = True
                    print(f'\t{d} is an IP address')
                except ValueError:
                    domains.add(d)
            else:
                error = True
                print(f'\t{d} is invalid')

    return list(sorted(domains))

def ask_ips() -> list[str]:
    """
    Ask the user for IP addresses.

    The user can enter 0, 1 or n IP addresses separated by a comma. IPv4 and IPv6 are supported.

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the IPs
    """
    error = True
    ips = set()

    while error:
        input_ips = ask('Enter IPs separated by a comma (eg: 10.0.0.1, 10.254.0.1)', can_be_empty=True)
        ips.clear()

        if not input_ips or len(input_ips) == 0:
            break

        error = False
        for ip in input_ips.split(','):
            ip = ip.strip()

            try:
                ipaddress.ip_address(ip)
                ips.add(ip)
            except ValueError:
                error = True
                print(f'\t{ip} is invalid')

    return list(sorted(ips))

def ask_cn_and_san() -> tuple[str, x509.SubjectAlternativeName]:
    """
    Ask the user for the CN, the alternative hosts and alternative IPS.

    :return: the CN and the build SAN
    """
    cn = ask_cn()
    alternate_hosts = ask_hosts()
    alternate_ips = ask_ips()

    try:  # CN is IP
        ipaddress.ip_address(cn)
        alternate_ips.append(cn)
    except ValueError:
        alternate_hosts.append(cn)

    san = [x509.DNSName(h) for h in alternate_hosts]
    san += [x509.IPAddress(ipaddress.ip_address(i)) for i in alternate_ips]

    return cn, x509.SubjectAlternativeName(san)

def confirm(prompt: str) -> bool:
    """
    Ask user to enter Y or N (case-insensitive) to confirm an action.

    :param prompt: the prompt message
    :return: True if the answer is Y.
    """
    return ask(f'{prompt} ? [y/n]', ['y', 'n'], disable_show_values=True) == 'y'

def ask_password(prompt: str) -> str:
    """
    Ask (securely) the user to enter a password.

    :param prompt: the prompt message
    :return: the password encoded
    """
    prompt += ' : '

    return getpass.getpass(prompt)

def ask_private_key_passphrase() -> str:
    """
    Ask the user for the private key passphrase.

    :return: the passphrase
    """
    return ask_password('Enter the private key passphrase')

def generate_random_serial_number() -> int:
    """
    Generate (securely) a random serial number for a certificate [0, 2^64].

    :return: the serial number
    """
    return secrets.randbelow(2 ** 64)

class HasExtensions(Protocol):
    """
    Protocol describing any object exposing X.509 extensions.

    This protocol is used for static duck typing: any object that provides
    an ``extensions`` attribute of type ``x509.Extensions`` is considered
    compatible, regardless of its concrete class.

    This allows writing generic functions operating on both
    ``x509.Certificate`` and ``x509.CertificateSigningRequest`` (and any
    future compatible type) without requiring a common base class.

    Notes:
       - This protocol is only used by static type checkers (mypy, pyright).
       - It has no effect at runtime and introduces no performance cost.
       - No inheritance is required from the implementing classes.
    """
    @property
    def extensions(self) -> x509.Extensions: ...

def get_san(obj: HasExtensions) -> Optional[x509.SubjectAlternativeName]:
    """
    Get the Subject Alternative Name from a certificate object.

    :param obj: the certificate or
    :return: the SAN if defined
    """
    try:
        extension = obj.extensions.get_extension_for_oid(
            ExtensionOID.SUBJECT_ALTERNATIVE_NAME
        )

        return cast(x509.SubjectAlternativeName, extension.value)
    except x509.ExtensionNotFound:
        pass

    return None

def get_san_str(san_extension_value: x509.SubjectAlternativeName) -> Optional[str]:
    """
    Get a string from the SAN extension value.

    A host is "DNS:<host>".
    An IP is "IP:<ip>"
    Each value is separated with a coma + space

    :param san_extension_value: the SAN extension value
    :return: the SAN if defined
    """
    san_list = []
    for name in san_extension_value:
        if isinstance(name, x509.DNSName):
            san_list.append(f'DNS:{name.value}')
        elif isinstance(name, x509.IPAddress):
            san_list.append(f'IP:{name.value}')
    return ', '.join(san_list) if san_list else None

def get_cn(o: x509.Certificate|x509.CertificateSigningRequest) -> str:
    """
    Get the Common Name from a certificate signing request object.

    :param o: the certificate or the certificate signing request
    :return: the CN
    """
    cn_attributes = o.subject.get_attributes_for_oid(NameOID.COMMON_NAME)

    if cn_attributes:
        return cn_attributes[0].value
    else:
        return ""

def get_crl_urls(cert: x509.Certificate) -> Optional[list[str]]:
    """
    Get the CRL URLs from a certificate.

    :param cert: the certificate
    :return: a list of URL or None
    """

    crl_urls = []
    try:
        ext = cert.extensions.get_extension_for_oid(ExtensionOID.CRL_DISTRIBUTION_POINTS)
        for dp in ext.value:
            for name in dp.full_name:
                if isinstance(name, x509.UniformResourceIdentifier):
                    crl_urls.append(name.value)
    except x509.ExtensionNotFound:
        pass

    if len(crl_urls) > 0:
        return crl_urls
    else:
        return None

def get_ocsp_url(cert: x509.Certificate) -> Optional[list[str]]:
    """
    Get the OCSP URLs from a certificate.

    :param cert: the certificate
    :return: a list of URL or None
    """
    ocsp_urls = []

    try:
        aia = cert.extensions.get_extension_for_oid(ExtensionOID.AUTHORITY_INFORMATION_ACCESS).value

        for access_desc in aia:
            if access_desc.access_method == AuthorityInformationAccessOID.OCSP:
                ocsp_urls.append(access_desc.access_location.value)
    except x509.ExtensionNotFound:
        pass

    if len(ocsp_urls) > 0:
        return ocsp_urls
    else:
        return None

def validate_or_build_san(cn: str, san: Optional[x509.SubjectAlternativeName]) -> x509.SubjectAlternativeName:
    """
    Ensure that the CN is in the SAN. Create SAN if empty.

    :param cn: the Common Name
    :param san: the optional Subject Alternative Name
    :return: the final Subject Alternative Name
    """
    # Checking if the CN is an IP address or a hostname
    try:
        cn_obj = x509.IPAddress(ipaddress.ip_address(cn))
    except ValueError:
        cn_obj = x509.DNSName(cn)

    #  Browse SAN
    san_entries = []
    cn_found_in_san = False

    if san is not None:
        for i in san:
            if i.value == cn_obj.value:
                cn_found_in_san = True

            san_entries.append(i)

    if cn_found_in_san:
        return san

    san_entries.append(cn_obj)

    return x509.SubjectAlternativeName(san_entries)

def get_cert_text(cert: x509.Certificate) -> str:
    """
    Get the text version of a certificate object.

    :param cert: the certificate
    :return: the text version
    """
    lines = [f"Certificate:"]

    lines.append(f"\tSerial Number: {cert.serial_number:x}")
    lines.append(f"\tSignature Algorithm: {cert.signature_algorithm_oid._name}")
    lines.append(f"\tIssuer: {cert.issuer.rfc4514_string()}")
    lines.append(f"\tValidity:")
    lines.append(f"\t\tNot Before: {cert.not_valid_before_utc}")
    lines.append(f"\t\tNot After : {cert.not_valid_after_utc}")
    lines.append(f"\tSubject: {cert.subject.rfc4514_string()}")

    # Extensions
    if len(cert.extensions):
        lines.append("\tX509v3 extensions:")
        lines.append(x509v3_extensions_to_str(cert.extensions))

    return "\n".join(lines)

def get_cert_builder_text(cert_builder: x509.CertificateBuilder) -> str:
    """
    Get the text version of a certificate builder object.

    :param cert_builder: the certificate builder
    :return: the text version
    """
    lines = [f"Certificate:"]

    lines.append(f"\tIssuer: {cert_builder._issuer_name.rfc4514_string()}")
    lines.append(f"\tValidity:")
    lines.append(f"\t\tNot Before: {cert_builder._not_valid_before}")
    lines.append(f"\t\tNot After : {cert_builder._not_valid_after}")
    lines.append(f"\tSubject: {cert_builder._subject_name.rfc4514_string()}")

    # Extensions
    if len(cert_builder._extensions):
        lines.append("\tX509v3 extensions:")
        lines.append(x509v3_extensions_to_str(cert_builder._extensions))

    return "\n".join(lines)

def get_csr_text(csr: x509.CertificateSigningRequest) -> str:
    """
    Get the text version of a certificate signing request object.

    :param csr: the certificate signing request
    :return: the text version
    """
    lines = [f"Certificate Request:"]

    lines.append(f"\tSignature Algorithm: {csr.signature_algorithm_oid._name}")
    lines.append(f"\tSubject: {csr.subject.rfc4514_string()}")

    # Extensions
    if len(csr.extensions):
        lines.append("\tX509v3 extensions:")
        lines.append(x509v3_extensions_to_str(csr.extensions))

    return "\n".join(lines)

def x509v3_extensions_to_str(extensions: x509.Extensions) -> str:
    """
    Convert the x509 v3 extensions to a string.

    :param extensions: the x509 v3 extensions
    :return: the text version, can be empty
    """
    lines = []
    for ext in extensions:
        name = ext.oid._name
        value = ext.value
        if isinstance(value, x509.SubjectAlternativeName):
            value = get_san_str(value)
        elif isinstance(value, x509.SubjectKeyIdentifier):
            value = binascii.hexlify(value.digest, sep=':').decode('utf8')
        elif isinstance(value, x509.AuthorityKeyIdentifier):
            value = binascii.hexlify(value.key_identifier, sep=':').decode('utf8')
        elif isinstance(value, x509.CRLDistributionPoints):
            values_str = []

            for v in value:
                for res in v.full_name:
                    values_str.append(res.value)

            value =  ', '.join(values_str)
        elif isinstance(value, x509.AuthorityInformationAccess):
            values_str = []

            for d in value:
                if isinstance(d, x509.AccessDescription):
                    if d.access_method == AuthorityInformationAccessOID.OCSP:
                        values_str.append(f'OCSP:{d.access_location.value}')
                else:
                    values_str.append(str(d))

            value = ', '.join(values_str)

        lines.append(f"\t\t{name}: {value}")


    return "\n".join(lines)

def nb_seconds_to_date(expire_date: datetime.date) -> int:
    """
    Compute the number of seconds between now and the expired date. Note that the date is given without the time part
    so we use the current hours/minutes/seconds.

    :param expire_date: the expiration date
    :return: the number of seconds
    """
    if expire_date < datetime.date.today():
        raise ValueError('The expire date is not in the future')

    expire_dt = datetime.datetime.now().replace(year=expire_date.year, month=expire_date.month, day=expire_date.day)
    return int((expire_dt - datetime.datetime.now()).total_seconds())

def print_tabulated(text: str) -> None:
    """
    Print the text with one tabulation.

    :param text: the text to print
    """

    text = '\t' + text.replace('\n', '\n\t')
    print(text)

def get_cert_pem(cert: x509.Certificate) -> str:
    """
    Get the certificate in PEM format.

    :param cert: the certificate
    :return: The PEM string
    """
    return cert.public_bytes(encoding=serialization.Encoding.PEM).decode('utf8')

def get_private_key_pem(key: PrivateKeyTypes) -> str:
    """
    Get the private key in PEM format

    :param key: the private key
    :return: the PEM string
    """
    return key.private_bytes(encoding=serialization.Encoding.PEM,
                             format=serialization.PrivateFormat.TraditionalOpenSSL,
                             encryption_algorithm=serialization.NoEncryption()).decode('utf8')

def write_cert_pem(cert: x509.Certificate, filepath: str):
    """
    Write the certificate in PEM format into the provided filepath.

    If the file already exists, a RuntimeError is raised.

    :param cert: the certificate
    :param filepath: the target filepath
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    with open(filepath, 'wb') as f:
        f.write(cert.public_bytes(encoding=serialization.Encoding.PEM))

def write_private_key_pem(key: PrivateKeyTypes, filepath: str):
    """
    Write the private key in PEM format into the provided filepath.

    If the file already exists, a RuntimeError is raised.

    :param key: the private key
    :param filepath: the target filepath
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    with open(filepath, 'wb') as f:
        f.write(
            get_private_key_pem(key).encode('utf8')
        )

def get_private_key_from_pem(pem: str) -> PrivateKeyTypes:
    """
    Get the private key from PEM format.

    Ask password if needed. Raise RuntimeError if the PEM is invalid or if the password is incorrect

    :param pem: the PEM string
    :return: the private key
    :raise RuntimeError: if the private key or the password is incorrect
    """
    try:
        try:
            key_obj = serialization.load_pem_private_key(pem.encode('utf8'),
                                                         password=None,
                                                         backend=default_backend())
        except TypeError:  # Key encrypted
            key_obj = serialization.load_pem_private_key(pem.encode('utf8'),
                                                         password=ask_private_key_passphrase().encode('utf8'),
                                                         backend=default_backend())
    except ValueError:
        raise RuntimeError('Invalid passphrase or invalid private key')

    return key_obj

def verify_signature(child: x509.Certificate, issuer: x509.Certificate) -> bool:
    """
    Verify that child is signed by issuer_public_key

    :param child: the child certificate
    :param issuer: the issuer certificate
    :return: bool
    """
    issuer_public_key = issuer.public_key()
    try:
        if isinstance(issuer_public_key, rsa.RSAPublicKey):
            issuer_public_key.verify(
                signature=child.signature,
                data=child.tbs_certificate_bytes,
                padding=padding.PKCS1v15(),
                algorithm=child.signature_hash_algorithm,
            )

        elif isinstance(issuer_public_key, ec.EllipticCurvePublicKey):
            issuer_public_key.verify(
                signature=child.signature,
                data=child.tbs_certificate_bytes,
                signature_algorithm=ec.ECDSA(child.signature_hash_algorithm),
            )
        else:
            raise TypeError("Unsupported public key type")

        return True
    except InvalidSignature:
        return False

def write_p12(filepath: str, passphrase: str, name: str, cert: x509.Certificate, key: Optional[PrivateKeyTypes] = None):
    """
    Write the certificate (and the private key if provided) in PKCS12 format.

    If the file already exists, a RuntimeError is raised.

    :param filepath: the target filepath
    :param passphrase: the password
    :param name: the certificate name
    :param cert: the certificate
    :param key: the private key
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    if passphrase is not None and len(passphrase) == 0:
        raise ValueError('The provided passphrase is empty')

    if passphrase:
        encryption = serialization.BestAvailableEncryption(passphrase.encode())
    else:
        encryption = serialization.NoEncryption()

    p12 = pkcs12.serialize_key_and_certificates(
        name=name.encode('utf8'),
        key=key,
        cert=cert,
        cas=None,
        encryption_algorithm=encryption
    )

    with open(filepath, "wb") as f:
        f.write(p12)

def decrypt(data: bytes, key: bytes, nonce: bytes, tag: Optional[bytes] = None) -> bytes:
    """
    Decrypt data with the key (AES_EAX).
    """
    cipher = AES.new(key, AES.MODE_GCM, nonce)

    if tag:
        return unpad(cipher.decrypt_and_verify(data, tag), 16)
    else:
        return unpad(cipher.decrypt(data), 16)

def decrypt_from_b64(data: str, key: bytes) -> bytes:
    """
    Decrypt data with the key (AES_EAX) from base64
    """
    data = base64.b64decode(data)

    return decrypt(data=data[32:], key=key, nonce=data[:16], tag=data[16:32])

def encrypt(plaintext: bytes, key: bytes) -> tuple[bytes, bytes, bytes]:
    """
    Encrypt data with the key (AES_EAX)
    """
    cipher = AES.new(key, AES.MODE_GCM)

    encrypted, tag = cipher.encrypt_and_digest(pad(plaintext, 16))
    return cipher.nonce, tag, encrypted

def encrypt_to_b64(plaintext: bytes, key: bytes) -> str:
    """
    Encrypt data with the key (AES_EAX) and return it encrypted as a base64 string
    """
    nonce, tag, encrypted = encrypt(plaintext, key)

    return base64.b64encode(nonce + tag + encrypted).decode('utf8')

def singleton(c):
    """
    Singleton decorator

    :param c: the class
    :return: the class instance
    """
    instances = {}

    def getinstance(*args, **kwargs):
        if c not in instances:
            instances[c] = c(*args, **kwargs)

        return instances[c]

    return getinstance

def is_valid_http_url(url: str) -> bool:
    """
    Check if a URL is a valid HTTP(S) URL.

    :param url: the URL
    :return: True or False
    """
    try:
        result = urlparse(url)
        return (
            result.scheme in ("http", "https") and
            bool(result.netloc)
        )
    except Exception:
        return False

def format_serial_number(cert: x509.Certificate) -> str:
    """
    Convert a certificate serial_number to OpenSSL-style hex with colons.

    :param cert: the certificate
    """
    serial = cert.serial_number  #
    hex_serial = f"{serial:x}"

    # Add "0" if needed
    if len(hex_serial) % 2 != 0:
        hex_serial = "0" + hex_serial

    # Regrouper par octets
    octets = [hex_serial[i:i+2] for i in range(0, len(hex_serial), 2)]

    return ":".join(octets)


import argparse
import os
import shutil
import sys

from .camanager import CAManager
from .models import DATABASE_FILENAME, DATABASE_FILENAME_BACKUP
from .utils import confirm

def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, prog='camanager')

    subparsers = parser.add_subparsers(dest='actions', help='The action')
    subparsers.add_parser('setup', help='Initialize the vault')
    subparsers.add_parser('add-ca', help='Add a CA')
    subparsers.add_parser('backup', help='Backup the vault')
    subparsers.add_parser('restore', help='Restore the vault from the backup')
    newcert_subparsers = subparsers.add_parser('newcert', help='Create a new certificate')
    sign_subparsers = subparsers.add_parser('sign', help='Sign a certificate')
    renew_subparsers = subparsers.add_parser('renew', help='Renew a certificate')
    list_subparsers = subparsers.add_parser('list', help='List certificates')
    export_subparsers = subparsers.add_parser('export', help='Export a certificate')
    revoke_subparsers = subparsers.add_parser('revoke', help='Revoke a certificate')
    crl_subparsers = subparsers.add_parser('crl', help='Generate the CRL and the index')
    add_subparsers = subparsers.add_parser('add', help='Add a certificate generated externally')
    subparsers.required = True

    newcert_subparsers.add_argument('--ca', default='', help='The issuer CA')

    sign_subparsers.add_argument('--ca', default='', help='The issuer CA')
    sign_subparsers.add_argument('csr_file', nargs='?', help='The CSF filepath')

    renew_subparsers.add_argument('--ca', default='', help='The issuer CA')
    renew_subparsers.add_argument('certificate', help='The certificate # or Common Name')

    list_group = list_subparsers.add_mutually_exclusive_group()
    list_group.add_argument('--ca', default='', help='The issuer CA')
    list_group.add_argument('--all', action='store_true', help='Also the revoked/renewed/expired certificates')
    list_group.add_argument('--soon-expired', action='store_true', help='Only the soon expired (< 1 month)')

    export_group = export_subparsers.add_mutually_exclusive_group(required=True)
    export_group.add_argument('--ca', default='', help='The issuer CA')
    export_group.add_argument('--pem', action='store_true', help='PEM output')
    export_group.add_argument('--p12', action='store_true', help='P12 output')
    export_subparsers.add_argument('--out', nargs='?', help='The output file')
    export_subparsers.add_argument('certificate', help='The certificate # or Common Name')

    revoke_subparsers.add_argument('--ca', default='', help='The issuer CA')
    revoke_subparsers.add_argument('certificate', help='The certificate # or Common Name')

    crl_subparsers.add_argument('--ca', default='', help='The issuer CA')

    add_subparsers.add_argument('--ca', default='', help='The issuer CA')
    add_subparsers.add_argument('pem_file', nargs='?', help='The certificate PEM filepath')

    args = parser.parse_args()

    cam = CAManager()

    try:
        if args.actions == 'setup':
            if os.path.isfile(DATABASE_FILENAME):
                msg = f'Error : the database "{DATABASE_FILENAME}" already exists.\n' \
                      f'You can remove this file BUT YOU ARE GOING TO LOST EVERYTHING RELATED TO THE CURRENT ' \
                      f'CERTIFICATE AUTHORITY. Be sure of the consequence of your action.'

                sys.stderr.write(msg)
                exit(-1)

            try:
                cam.create_vault()
            except ValueError as e:
                sys.stderr.write(f'Error: {str(e)}\n')
                exit(-1)

        elif args.actions == 'add-ca':
            if not os.path.isfile(DATABASE_FILENAME):
                msg = f'Error : the database "{DATABASE_FILENAME}" not found.\n'

                sys.stderr.write(msg)
                exit(-1)

            try:
                cam.load(dont_load_ca=True)
                cam.add_ca()
            except ValueError as e:
                sys.stderr.write(f'Error: {str(e)}\n')
                exit(-1)

        elif args.actions == 'backup':
            if not os.path.isfile(DATABASE_FILENAME):
                msg = f'Error : the database "{DATABASE_FILENAME}" not found.\n'

                sys.stderr.write(msg)
                exit(-1)

            if os.path.isfile(DATABASE_FILENAME_BACKUP):
                msg = f'Error : the database backup "{DATABASE_FILENAME_BACKUP}" already exists.\n\n' \
                      f'You can remove this file BUT YOU ARE GOING TO LOST EVERYTHING RELATED TO THE BACKUP.'

                print(msg)

                if not confirm('Do you to ERASE the previous backup'):
                    print('No backup was done')
                    exit(0)
                else:
                    os.remove(DATABASE_FILENAME_BACKUP)

            shutil.copy(DATABASE_FILENAME, DATABASE_FILENAME_BACKUP)
            print('Backup done successfully')

        elif args.actions == 'restore':
            if not os.path.isfile(DATABASE_FILENAME_BACKUP):
                sys.stderr.write(f'No backup found')
                exit(-1)

            if os.path.isfile(DATABASE_FILENAME):
                msg = f'Warning : YOU ARE GOING TO LOST EVERYTHING RELATED TO THE CURRENT CERTIFICATE AUTHORITY. ' \
                      f'Be sure of the consequence of your action.\n'

                print(msg)

                if not confirm('Do you you want to restore the backup'):
                    print('The backup has NOT been restored')
                    exit(0)
                else:
                    os.remove(DATABASE_FILENAME)

            shutil.copy(DATABASE_FILENAME_BACKUP, DATABASE_FILENAME)

            print('Backup restored successfully')

            if not confirm('Do you want to keep the backup'):
                os.remove(DATABASE_FILENAME_BACKUP)
                print('Backup file deleted successfully')

        elif args.actions == 'newcert':
            cam.load(ca_name=args.ca)
            cam.generate_new_cert()

        elif args.actions == 'sign':
            cam.load(ca_name=args.ca)
            cam.sign_csr(args.csr_file)

        elif args.actions == 'renew':
            cam.load(ca_name=args.ca)
            cam.renew(args.certificate)

        elif args.actions == 'list':
            cam.load(ca_name=args.ca)
            cam.list(args.all, args.soon_expired)

        elif args.actions == 'export':
            cam.load(ca_name=args.ca)
            cam.export(args.certificate, "p12" if args.p12 else "pem", args.out)

        elif args.actions == 'revoke':
            cam.load(ca_name=args.ca)
            cam.revoke(args.certificate)

        elif args.actions == 'crl':
            cam.load(ca_name=args.ca)
            cam.generate_crl()

        elif args.actions == 'add':
            cam.load(ca_name=args.ca)
            cam.add_external(args.pem_file)



    except ValueError as e:
        sys.stderr.write(f'Error: {str(e)}\n')
        exit(-1)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print('\n\n*** Interrupted by user ***')



import datetime
import enum

import peewee
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.hashes import HashAlgorithm

from .masterkey import MasterKeyHelper
from .utils import get_private_key_from_pem

"""
The database filename
"""
DATABASE_FILENAME = 'ca.db'

"""
The database filename backup
"""
DATABASE_FILENAME_BACKUP = f'{DATABASE_FILENAME}.bak'

"""
The database object
"""
database = peewee.SqliteDatabase(None)

class FixedDateTimeField(peewee.DateTimeField):
    def python_value(self, value):
        if value is None:
            return None
        if isinstance(value, datetime.datetime):
            return value
        return datetime.datetime.fromisoformat(value)

class BaseModel(peewee.Model):
    """
    Base model
    """
    class Meta:
        database = database


class BaseCertKeyModel(BaseModel):
    """
    Base certificate model which may have a key.

    The certificate must be stored into the "cert" attribute in the PEM format.
    The optional private key must be stored into the "key" attribute in the PEM format.

    The x509.Certificate and PrivateKeyTypes can be loaded as self.cert_obj and self.key_obj. If the key is
    passphrase protected, the user is asking for this passphrase.
    """
    cert = peewee.TextField()
    encrypted_key = peewee.TextField(null=True)
    _key = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cert_obj = None
        self.key_obj = None

    def load_cert(self) -> None:
        """
        Load the certificate as x509.Certificate object into the attribute "cert_obj".
        """
        self.cert_obj = x509.load_pem_x509_certificate(str(self.cert).encode('utf8'), default_backend())

    def load_key(self) -> None:
        """
        Load the private key as PrivateKeyTypes object into the attribute "key_obj".

        If the key is passphrase protected, the user is asking for this passphrase (prompt to stdout, read from stdin).

        If no key is available (self.key is None), a RuntimeError is raised.

        :except: RunTime
        """
        if not self.key:
            raise RuntimeError('No key available')

        self.key_obj = get_private_key_from_pem(str(self.key))


    def clear_key(self) -> None:
        """
        Clear the attribute "key_obj".
        """
        self.key_obj = None

    def _get_key(self):
        if self.encrypted_key is None:
            return None

        return MasterKeyHelper().decrypt_from_b64(self.encrypted_key).decode('utf8')

    def _set_key(self, data: bytes):
        if data is None:
            self.encrypted_key = None
        else:
            if isinstance(data, str):
                data = data.encode('utf8')

            self.encrypted_key = MasterKeyHelper().encrypt_to_b64(data)

    key = property(fget=_get_key, fset=_set_key, doc="Private Key (PEM)")


class CRLSigningCertificate(BaseModel):
    """
    A CRL Signing Certificate.

    The certificate and the key are stored in the PEM format, in plain text !
    """
    cert = peewee.TextField()
    key = peewee.TextField()


class CA(BaseCertKeyModel):
    """
    A Certificate Authority model.

    The certificate and the key are stored in the PEM format.
    """
    id = peewee.IntegerField(primary_key=True)
    is_default = peewee.BooleanField(default=False, unique=True)
    name = peewee.CharField(unique=True)
    is_intermediate = peewee.BooleanField()
    crl_signing = peewee.ForeignKeyField(CRLSigningCertificate, backref='ca', null=True)
    crl_last_generated = FixedDateTimeField(null=True)
    crl_output_filename = peewee.CharField(unique=True)
    crl_url = peewee.TextField(null=True)
    ocsp_url = peewee.TextField(null=True)
    crl_post_script_path = peewee.TextField(null=True)


class CertificateRevokedReason(enum.Enum):
    """
    The certificate revoked reason
    """
    UNSPECIFIED = 'unspecified'
    KEY_COMPRISE = 'keyCompromise'
    CESSATION_OF_OPERATION = 'cessationOfOperation'

    @staticmethod
    def from_string(s: str):
        s = s.lower()

        if s in ['keycompromise', 'key compromise', 'compromise', 'compromised']:
            return CertificateRevokedReason.KEY_COMPRISE
        elif s in ['cessationOfOperation', 'cessation of operation', 'eol']:
            return CertificateRevokedReason.CESSATION_OF_OPERATION
        else:
            return CertificateRevokedReason.UNSPECIFIED

    def to_reason_flags(self) -> x509.ReasonFlags:
        match self:
            case self.KEY_COMPRISE:
                return x509.ReasonFlags.key_compromise
            case self.CESSATION_OF_OPERATION:
                return x509.ReasonFlags.cessation_of_operation
            case _:
                return x509.ReasonFlags.unspecified


class Certificate(BaseCertKeyModel):
    """
    A certificate model.

    The created_timestamp and not_after are precise to the second.
    The serial is provided in the official X509 output, so in hex format using ":" separator in the big endian
    representation of the 64 bits integer.
    The certificate and the key (optional) are stored in the PEM format.
    """
    id = peewee.IntegerField(primary_key=True)
    ca = peewee.ForeignKeyField(CA, backref='ca')
    cn = peewee.CharField()
    san = peewee.CharField(null=True)
    created_timestamp = FixedDateTimeField()
    not_after = FixedDateTimeField()
    serial = peewee.IntegerField()
    is_revoked = peewee.BooleanField()
    revoked_timestamp = FixedDateTimeField(null=True)
    revoked_reason = peewee.TextField(null=True)
    revoked_comment = peewee.TextField(null=True)
    is_renewed = peewee.BooleanField()

    @property
    def serial_as_int64(self):
        return int.from_bytes(bytes.fromhex(self.serial.replace(':', '')), 'big')

    def __str__(self):
        now = datetime.datetime.now(datetime.UTC)

        s = f'#{self.id}. CN:{self.cn}'\

        if self.san:
            s += f' / {self.san}'

        s += ' ('

        if self.not_after <= now:
            s += 'expired'
        elif self.is_revoked:
            s += 'revoked'

            if self.revoked_comment:
                s += f' [comment: {self.revoked_comment}]'
        elif self.is_renewed:
            s += 'renewed [must not be used anymore]'
        else:
            s += f'expire on {self.not_after.strftime("%d/%m/%Y %H:%M:%S %Z")}'

        s += f', serial: {self.serial}, '
        s += 'no private key stored' if not self.encrypted_key else 'private key available'
        s += ')'

        return s


class ConfigType(enum.IntEnum):
    """
    The configuration value type
    """
    INT = 1
    STRING = 2
    BINARY = 3
    EPOCH = 4


class Config(BaseModel):
    """
    A Configuration model
    """
    name = peewee.CharField(primary_key=True)
    type = peewee.IntegerField()
    value = peewee.CharField()

class HashAlgo(enum.Enum):
    SHA1 = hashes.SHA1()
    SHA256 = hashes.SHA256()
    SHA512 = hashes.SHA512()

    @staticmethod
    def from_string(s: str) -> "HashAlgo":
        match s.lower():
            case 'sha1':
                return HashAlgo.SHA1
            case 'sha256':
                return HashAlgo.SHA256
            case 'sha512':
                return HashAlgo.SHA512
            case _:
                raise NotImplementedError

    @staticmethod
    def from_x509_hash_algorithm(h: HashAlgorithm):
        return HashAlgo.from_string(h.name)

    @staticmethod
    def get_str_values():
        return ['sha1', 'sha256', 'sha512']



import binascii
import datetime
import unittest

from cryptography.hazmat.primitives import hashes

from src.camanager.models import CertificateRevokedReason, Certificate, FixedDateTimeField, HashAlgo, BaseCertKeyModel
from src.camanager.masterkey import MasterKeyHelper
from .data import *


class TestFixedDateTimeField(unittest.TestCase):
    def setUp(self):
        """Initialize the field before each test"""
        self.field = FixedDateTimeField()

    def test_python_value_with_none(self):
        """Test that None returns None"""
        result = self.field.python_value(None)
        self.assertIsNone(result)

    def test_python_value_with_datetime_object(self):
        """Test that a datetime object is returned as-is"""
        dt = datetime.datetime(2025, 1, 4, 12, 30, 45)
        result = self.field.python_value(dt)
        self.assertEqual(result, dt)
        self.assertIsInstance(result, datetime.datetime)

    def test_python_value_with_iso_string(self):
        """Test conversion from ISO string to datetime"""
        iso_string = "2025-01-04T12:30:45"
        result = self.field.python_value(iso_string)
        expected = datetime.datetime(2025, 1, 4, 12, 30, 45)
        self.assertEqual(result, expected)
        self.assertIsInstance(result, datetime.datetime)

    def test_python_value_with_iso_string_with_microseconds(self):
        """Test with microseconds"""
        iso_string = "2025-01-04T12:30:45.123456"
        result = self.field.python_value(iso_string)
        expected = datetime.datetime(2025, 1, 4, 12, 30, 45, 123456)
        self.assertEqual(result, expected)

    def test_python_value_with_iso_string_with_timezone(self):
        """Test with timezone"""
        iso_string = "2025-01-04T12:30:45+01:00"
        result = self.field.python_value(iso_string)
        self.assertIsInstance(result, datetime.datetime)
        self.assertEqual(result.hour, 12)

    def test_python_value_with_invalid_string(self):
        """Test that an invalid string raises an exception"""
        with self.assertRaises(ValueError):
            self.field.python_value("invalid-date-string")


class TestBaseCertKeyModel(unittest.TestCase):
    def setUp(self):
        helper = MasterKeyHelper()
        helper.masterkey = (b'\xf8\x48\x0f\xd8\xe7\xcd\xf0\x86\xfe\x12\x5e\x5a\xd0\x6f\x14\xc3'
                            b'\xa2\x44\xe2\x4c\xee\x10\x15\xcd\x58\x12\x52\xd6\x04\x11\xcc\x1a')


    def test_cert_obj(self):
        base_certkey = BaseCertKeyModel()
        base_certkey.cert = load_txt('tests/data/cert.pem')
        base_certkey.load_cert()

        self.assertEqual(base_certkey.cert_obj, get_cert_obj())

    def test_no_key(self):
        base_certkey = BaseCertKeyModel()
        base_certkey.cert = load_txt('tests/data/cert.pem')

        with self.assertRaises(RuntimeError):
            base_certkey.load_key()

        self.assertIsNone(base_certkey.key)

    def test_set_key(self):
        base_certkey = BaseCertKeyModel()
        base_certkey.cert = load_txt('tests/data/cert.pem')

        self.assertIsNone(base_certkey.key)
        self.assertIsNone(base_certkey.encrypted_key)

        base_certkey.key = load_txt('tests/data/pkey.key')

        self.assertIsNotNone(base_certkey.key)
        self.assertIsNotNone(base_certkey.encrypted_key)

        base_certkey.load_key()
        self.assertEqual(base_certkey.key_obj.private_bytes(serialization.Encoding.PEM,
                                                            serialization.PrivateFormat.TraditionalOpenSSL,
                                                            serialization.NoEncryption()),
                         get_privatekey_obj().private_bytes(serialization.Encoding.PEM,
                                                            serialization.PrivateFormat.TraditionalOpenSSL,
                                                            serialization.NoEncryption()))

        base_certkey.clear_key()
        self.assertIsNone(base_certkey.key_obj)
        self.assertIsNotNone(base_certkey.key)
        self.assertIsNotNone(base_certkey.encrypted_key)

        base_certkey.key = None

        self.assertIsNone(base_certkey.key_obj)
        self.assertIsNone(base_certkey.key)
        self.assertIsNone(base_certkey.encrypted_key)

class TestCertificateRevokedReason(unittest.TestCase):
    def test_from_string(self):
        self.assertEqual(CertificateRevokedReason.from_string('key compromise'), CertificateRevokedReason.KEY_COMPRISE)
        self.assertEqual(CertificateRevokedReason.from_string('compromise'), CertificateRevokedReason.KEY_COMPRISE)
        self.assertEqual(CertificateRevokedReason.from_string('eol'), CertificateRevokedReason.CESSATION_OF_OPERATION)

        self.assertEqual(CertificateRevokedReason.from_string('other'), CertificateRevokedReason.UNSPECIFIED)
        self.assertEqual(CertificateRevokedReason.from_string('key'), CertificateRevokedReason.UNSPECIFIED)

    def test_to_reason_flags(self):
        self.assertEqual(CertificateRevokedReason.KEY_COMPRISE.to_reason_flags(), x509.ReasonFlags.key_compromise)
        self.assertEqual(CertificateRevokedReason.CESSATION_OF_OPERATION.to_reason_flags(), x509.ReasonFlags.cessation_of_operation)
        self.assertEqual(CertificateRevokedReason.UNSPECIFIED.to_reason_flags(), x509.ReasonFlags.unspecified)


class TestCertificate(unittest.TestCase):
    def test_serial_as_int64(self):
        c = Certificate(serial='6b:8c:02:7a:8a:64:11:d0:e0:94:e9:22:58:99:1a:d6:95:58:0e:0e')
        self.assertEqual(613984332728918798661844833697261882337094274574, c.serial_as_int64)

        c = Certificate(serial='06:9a:8b:0b:aa:2a:53:a8')
        self.assertEqual(475845592856810408, c.serial_as_int64)

        c = Certificate(serial='28:11:59:2f:b8:5e:ed:be:0e:98:4c:e9:c6:2e:21:b1:27:bf:5a:17')
        self.assertEqual(228746512733118622626861842270018213062574561815, c.serial_as_int64)


    def test_str(self):
        now = datetime.datetime.now(datetime.UTC)
        expire = now + datetime.timedelta(days=365)

        c = Certificate(id=5,
                        cn='test.example.com',
                        created_timestamp=now,
                        not_after=expire,
                        serial='01:02:03:04:05:06:07:09')
        s = f'#5. CN:test.example.com (expire on {expire.strftime('%d/%m/%Y %H:%M:%S UTC')}, serial: 01:02:03:04:05:06:07:09, no private key stored)'
        self.assertEqual(s, str(c))

        c = Certificate(id=5,
                        cn='test.example.com',
                        san='DNS:test.example.com',
                        created_timestamp=now,
                        not_after=expire,
                        serial='01:02:03:04:05:06:07:09',
                        is_revoked=True,
                        is_renewed=False)
        s = f'#5. CN:test.example.com / DNS:test.example.com (revoked, serial: 01:02:03:04:05:06:07:09, no private key stored)'
        self.assertEqual(s, str(c))

        c = Certificate(id=5,
                        cn='test.example.com',
                        san='DNS:test.example.com',
                        created_timestamp=now,
                        not_after=expire,
                        serial='01:02:03:04:05:06:07:09',
                        is_revoked=True,
                        revoked_reason=CertificateRevokedReason.KEY_COMPRISE,
                        revoked_comment='test comment',
                        is_renewed=False)
        s = f'#5. CN:test.example.com / DNS:test.example.com (revoked [comment: test comment], serial: 01:02:03:04:05:06:07:09, no private key stored)'
        self.assertEqual(s, str(c))

        c = Certificate(id=5,
                        cn='test.example.com',
                        san='DNS:test.example.com',
                        created_timestamp=now,
                        not_after=expire,
                        serial='01:02:03:04:05:06:07:09',
                        is_revoked=False,
                        is_renewed=True)
        s = f'#5. CN:test.example.com / DNS:test.example.com (renewed [must not be used anymore], serial: 01:02:03:04:05:06:07:09, no private key stored)'
        self.assertEqual(s, str(c))

        c = Certificate(id=5,
                        cn='test.example.com',
                        created_timestamp=now,
                        not_after=now - datetime.timedelta(days=1),
                        serial='01:02:03:04:05:06:07:09')
        s = f'#5. CN:test.example.com (expired, serial: 01:02:03:04:05:06:07:09, no private key stored)'
        self.assertEqual(s, str(c))


class TestHashAlgo(unittest.TestCase):
    def test_from_string(self):
        self.assertEqual(HashAlgo.from_string('sha1'), HashAlgo.SHA1)
        self.assertEqual(HashAlgo.from_string('SHA256'), HashAlgo.SHA256)
        self.assertEqual(HashAlgo.from_string('SHA512'), HashAlgo.SHA512)

        with self.assertRaises(NotImplementedError):
            HashAlgo.from_string('SHA 512')

    def test_from_x509_hash_algorithm(self):
        self.assertEqual(HashAlgo.from_x509_hash_algorithm(hashes.SHA1()), HashAlgo.SHA1)

    def test_get_str_values(self):
        self.assertEqual(['sha1', 'sha256', 'sha512'], HashAlgo.get_str_values())
